from os import PathLike
from unittest.mock import patch

import pytest

from gso.products import Iptrunk, ProductType
from gso.products.product_blocks.iptrunk import IptrunkType
from gso.services.crm import customer_selector, get_customer_by_name
from gso.services.subscriptions import get_product_id_by_name
from gso.utils.types.phy_port import PhyPortCapacity
from test.workflows import (
    assert_aborted,
    assert_complete,
    assert_suspended,
    extract_state,
    resume_workflow,
    run_workflow,
    user_accept_and_assert_suspended,
)


class MockedNetboxClient:
    class BaseMockObject:
        def __init__(self, **kwargs):
            for key, value in kwargs.items():
                setattr(self, key, value)

    def get_device_by_name(self):
        return self.BaseMockObject(id=1, name="test")

    def get_available_lags(self) -> list[str]:
        return [f"LAG{lag}" for lag in range(1, 5)]

    def get_available_interfaces(self):
        interfaces = []
        for interface in range(1, 5):
            interface_data = {
                "name": f"Interface{interface}",
                "module": {"display": f"Module{interface}"},
                "description": f"Description{interface}",
            }
            interfaces.append(interface_data)
        return interfaces

    def create_interface(self):
        return self.BaseMockObject(id=1, name="test")

    def attach_interface_to_lag(self):
        return self.BaseMockObject(id=1, name="test")

    def reserve_interface(self):
        return self.BaseMockObject(id=1, name="test")

    def allocate_interface(self):
        return {"id": 1, "name": "test"}


@pytest.fixture
def netbox_client_mock():
    # Mock NetboxClient methods
    with (
        patch("gso.services.netbox_client.NetboxClient.get_device_by_name") as mock_get_device_by_name,
        patch("gso.services.netbox_client.NetboxClient.get_available_interfaces") as mock_get_available_interfaces,
        patch("gso.services.netbox_client.NetboxClient.get_available_lags") as mock_get_available_lags,
        patch("gso.services.netbox_client.NetboxClient.create_interface") as mock_create_interface,
        patch("gso.services.netbox_client.NetboxClient.attach_interface_to_lag") as mock_attach_interface_to_lag,
        patch("gso.services.netbox_client.NetboxClient.reserve_interface") as mock_reserve_interface,
        patch("gso.services.netbox_client.NetboxClient.allocate_interface") as mock_allocate_interface,
    ):
        mock_get_device_by_name.return_value = MockedNetboxClient().get_device_by_name()
        mock_get_available_interfaces.return_value = MockedNetboxClient().get_available_interfaces()
        mock_get_available_lags.return_value = MockedNetboxClient().get_available_lags()
        mock_create_interface.return_value = MockedNetboxClient().create_interface()
        mock_attach_interface_to_lag.return_value = MockedNetboxClient().attach_interface_to_lag()
        mock_reserve_interface.return_value = MockedNetboxClient().reserve_interface()
        mock_allocate_interface.return_value = MockedNetboxClient().allocate_interface()

        yield


@pytest.fixture
def input_form_wizard_data(router_subscription_factory, faker):
    router_side_a = router_subscription_factory()
    router_side_b = router_subscription_factory()

    create_ip_trunk_step = {
        "tt_number": faker.pystr(),
        "customer": getattr(customer_selector(), get_customer_by_name("GÉANT")["id"]),
        "geant_s_sid": faker.pystr(),
        "iptrunk_type": IptrunkType.DARK_FIBER,
        "iptrunk_description": faker.sentence(),
        "iptrunk_speed": PhyPortCapacity.HUNDRED_GIGABIT_PER_SECOND,
        "iptrunk_minimum_links": 2,
    }
    create_ip_trunk_side_a_router_name = {"iptrunk_sideA_node_id": router_side_a}
    create_ip_trunk_side_a_step = {
        "iptrunk_sideA_ae_iface": "LAG1",
        "iptrunk_sideA_ae_geant_a_sid": faker.pystr(),
        "iptrunk_sideA_ae_members": ["Interface1", "Interface2"],
        "iptrunk_sideA_ae_members_descriptions": ["Interface1 Description", "Interface2 Description"],
    }

    create_ip_trunk_side_b_router_name = {"iptrunk_sideB_node_id": router_side_b}
    create_ip_trunk_side_b_step = {
        "iptrunk_sideB_ae_iface": "LAG1",
        "iptrunk_sideB_ae_geant_a_sid": faker.pystr(),
        "iptrunk_sideB_ae_members": ["Interface1", "Interface2"],
        "iptrunk_sideB_ae_members_descriptions": ["Interface1 Description", "Interface2 Description"],
    }

    return [
        create_ip_trunk_step,
        create_ip_trunk_side_a_router_name,
        create_ip_trunk_side_a_step,
        create_ip_trunk_side_b_router_name,
        create_ip_trunk_side_b_step,
    ]


@pytest.mark.workflow
@patch("gso.workflows.iptrunk.create_iptrunk.provisioning_proxy.check_ip_trunk")
@patch("gso.workflows.iptrunk.create_iptrunk.provisioning_proxy.provision_ip_trunk")
@patch("gso.workflows.iptrunk.create_iptrunk.infoblox.allocate_v6_network")
@patch("gso.workflows.iptrunk.create_iptrunk.infoblox.allocate_v4_network")
def test_successful_iptrunk_creation_with_standard_lso_result(
    mock_allocate_v4_network,
    mock_allocate_v6_network,
    mock_provision_ip_trunk,
    mock_check_ip_trunk,
    responses,
    input_form_wizard_data,
    faker,
    data_config_filename: PathLike,
    netbox_client_mock,
):
    mock_allocate_v4_network.return_value = faker.ipv4_network()
    mock_allocate_v6_network.return_value = faker.ipv6_network()
    product_id = get_product_id_by_name(ProductType.IP_TRUNK)
    initial_site_data = [{"product": product_id}, *input_form_wizard_data]
    result, process_stat, step_log = run_workflow("create_iptrunk", initial_site_data)
    assert_suspended(result)

    standard_lso_result = {
        "pp_run_results": {
            "status": "ok",
            "job_id": "random_job_id",
            "output": "parsed_output",
            "return_code": 0,
        },
        "confirm": "ACCEPTED",
    }
    for _ in range(5):
        result, step_log = user_accept_and_assert_suspended(process_stat, step_log, standard_lso_result)
        result, step_log = user_accept_and_assert_suspended(process_stat, step_log, [{}, {}])

    result, step_log = user_accept_and_assert_suspended(process_stat, step_log, standard_lso_result)
    result, step_log = resume_workflow(process_stat, step_log, [{}, {}])
    assert_complete(result)

    state = extract_state(result)
    subscription_id = state["subscription_id"]
    subscription = Iptrunk.from_subscription(subscription_id)

    assert "active" == subscription.status
    assert subscription.description == f"IP trunk, geant_s_sid:{input_form_wizard_data[0]['geant_s_sid']}"

    assert mock_provision_ip_trunk.call_count == 4
    assert mock_check_ip_trunk.call_count == 2


@pytest.mark.workflow
@patch("gso.workflows.iptrunk.create_iptrunk.provisioning_proxy.check_ip_trunk")
@patch("gso.workflows.iptrunk.create_iptrunk.provisioning_proxy.provision_ip_trunk")
@patch("gso.workflows.iptrunk.create_iptrunk.infoblox.allocate_v6_network")
@patch("gso.workflows.iptrunk.create_iptrunk.infoblox.allocate_v4_network")
def test_iptrunk_creation_fails_when_lso_return_code_is_one(
    mock_allocate_v4_network,
    mock_allocate_v6_network,
    mock_provision_ip_trunk,
    mock_check_ip_trunk,
    responses,
    input_form_wizard_data,
    faker,
    netbox_client_mock,
    data_config_filename: PathLike,
):
    mock_allocate_v4_network.return_value = faker.ipv4_network()
    mock_allocate_v6_network.return_value = faker.ipv6_network()
    product_id = get_product_id_by_name(ProductType.IP_TRUNK)

    initial_site_data = [{"product": product_id}, *input_form_wizard_data]
    result, process_stat, step_log = run_workflow("create_iptrunk", initial_site_data)
    assert_suspended(result)

    standard_lso_result = {
        "pp_run_results": {
            "status": "ok",
            "job_id": "random_job_id",
            "output": "parsed_output",
            "return_code": 1,
        },
        "confirm": "ACCEPTED",
    }

    attempts = 3
    for _ in range(0, attempts - 1):
        result, step_log = user_accept_and_assert_suspended(process_stat, step_log, standard_lso_result)
        result, step_log = user_accept_and_assert_suspended(process_stat, step_log, [{}, {}])

    result, step_log = user_accept_and_assert_suspended(process_stat, step_log, standard_lso_result)
    result, step_log = resume_workflow(process_stat, step_log, [{}, {}])
    assert_aborted(result)

    assert mock_provision_ip_trunk.call_count == attempts
    assert mock_check_ip_trunk.call_count == 0
